import numpy as np
from itertools import compress
from .field import *
from .quasiparticles import *
from scipy.interpolate import RegularGridInterpolator 
from .charge_propagation import *
from tqdm import tqdm
import csv
from .silicon import *

#from line_profiler import LineProfiler

#from joblib import Parallel, delayed
#import multiprocessing

#num_cores = multiprocessing.cpu_count()

class Simulation:
    '''
    Object contains all nessie objects needed to simulate a signal. 
    Currently assumes only one contact.
    
    '''
    def __init__(self, _name, _electricField=None, _weightingPotential=None, _electricPotential=None, _weightingField=None,  
                    _cceField=None, _chargeCaptureField=None, _electronicResponse=None, _temp=None):        
        self.name = _name
        self.electricField = _electricField
        self.electricPotential = _electricPotential
        self.weightingPotential = _weightingPotential
        self.weightingField = _weightingField
        self.cceField = _cceField
        self.chargeCaptureField = _chargeCaptureField
        self.electronicResponse = _electronicResponse
        self.temp = _temp
        self.bounds = None
        
    def __str__(self):
       return ("Nessie Simulation object \n Name: %s \n Electric Field: %s \n Weighting Potential: %s \n CCE Field: %s \n Charge Capture Field: %s \n Electronic Response: %s \n"
                % (self.name, self.electricField,self.weightingPotential, self.cceField, self.chargeCaptureField,          
                   self.electronicResponse))
    
    def setTemp(self,T):
        self.temp = T
        return None
        
    def setBounds(self, bounds):
        self.bounds = bounds
        return None
    
    def setElectricField(self):
        grid = self.electricPotential.grid
        gradient = np.gradient(self.electricPotential.data, grid[0],grid[1],grid[2])
        try:
            self.electricField = Field("Electric Field", gradient[0],gradient[1],gradient[2], self.electricPotential.gri)
        except:
            print("No electric potential from which to calculate weighting field.")
        return None

    def setWeightingField(self):
        grid = self.weightingPotential.grid
        gradient = np.gradient(self.weightingPotential.data, grid[0],grid[1],grid[2])
        try:
            self.weightingField = Field("Weighting Field", gradient[0],gradient[1],gradient[2], self.weightingPotential.grid)
        except:
            print("No weighting potential from which to calculate weighting field.")
        return None

    def setChargeCollectionEfficiencyField(self):
        return None

    def setChargeCaptureField(self):
        return None

    def setElectronicResponse(self, spiceFile):
        ts = []
        step = []
        with open(spiceFile,'r') as csvfile:
            plots = csv.reader(csvfile, delimiter=',')
            for row in plots:
                ts.append(float(row[0])*1e-9) #convert from ns to s
                step.append(float(row[1]))
        self.electronicResponse = {"times":ts, "step":step}
        return None

    def simulate(self, events, eps, dt, plasma=False, diffusion=False, capture=False, d=None, interp3d = True, maxPairs=100):
        '''
        Where it all happens! 
        When calling this function you determine which effects you want to simulate (eg. plasma, diffusion, etc.)
        Reduce maxPairs to run faster by having fewer quasiparticles with greater charge.
        Setting interp3d to true here will use our cython interpolator throughout the simulation.
        For machines with many cores you can try setting parallel=True. We see slight speedups but your mileage may vary. 
        '''
    
        # Get electric field interpolations
        Ex_i, Ey_i, Ez_i, Emag_i = self.electricField.interpolate(interp3d)
        
        simBounds = self.bounds if self.bounds is not None else [[axis[0],axis[-1]] for axis in self.electricField.grid]

        #Find electron and hole drift paths for each event
        for i in tqdm(range(len(events))):
            event = events[i]
            #rint will give an integer number of e-h pairs, but will need to use Fano factor for a proper calculation
            pairs = np.rint(event.dE/ephBestFit(self.temp))

            # electron charge cloud assembly
            cc_e = []
            # hole charge cloud assembly
            cc_h = []

            # Add charge cloud parts at each position where energy was deposited
            for j in tqdm(range(len(event.pos))):
                pairNr = int(pairs[j])
                factor = 1
                if pairNr > maxPairs:
                    factor = pairNr/maxPairs
                    pairNr = maxPairs

                # TODO: Provide a radius to smooth initial charge cloud (currently 0)
                cc_e = cc_e + initializeChargeCloud(-factor*qe_SI, factor*me_SI, pairNr, event.times[j], 0, event.pos[j])
                cc_h = cc_h + initializeChargeCloud(factor*qe_SI, factor*me_SI, pairNr, event.times[j], 0, event.pos[j])

            cc = cc_e + cc_h

            alive = np.ones(len(cc)) == 1
            print("Total quasiparticles: %d" % len(cc))

            #lp = LineProfiler()
            #lp.add_function(generalized_mobility_el)
            #lp_wrapper = lp(updateQuasiParticles)

            # Loop over alive particles until all have been stopped/collected
            pbar = tqdm()
            while np.any(alive):
                cc_new = updateQuasiParticles(list(compress(cc, alive)), eps, dt, Ex_i, Ey_i, Ez_i, Emag_i, simBounds, self.temp, diffusion=diffusion, coulomb=plasma)
                #cc_new = lp_wrapper(list(compress(cc, alive)), eps, dt, Ex_i, Ey_i, Ez_i, Emag_i, simBounds, self.temp, diffusion=diffusion, coulomb=plasma)
                counter = 0
                for j in range(len(alive)):
                    if alive[j]:
                        cc[j] = cc_new[counter]
                        counter+=1

                alive = np.array([o.alive for o in cc])
                pbar.update(1)

            event.quasiparticles = cc
            #lp.print_stats()
        return events

    def calculateInducedCurrent(self, events, dt):
        for event in events:
            event.calculateInducedCurrent(self.weightingField, dt, interp3d=True)

    def calculateElectronicResponse(self, events):
        for event in events:
            event.convolveElectronicResponse(self.electronicResponse)


        '''#get induced current
        print("calculating induced current")
        
        if self.weightingField is None: self.setWeightingField()
        
        for event in events:
            event.calculateInducedCurrent(self.weightingField, 0.1e-9, interp3d=interp3d)
            
        if self.electronicResponse is not None:
                for event in events:
                    event.convolveElectronicResponse(self.electronicResponse)'''




import numpy as np
from scipy.stats import maxwell

class Quasiparticle:
    '''
    Container object for a generalized quasiparticle with arbitrary mass and charge.
    Tracks position and velocity.
    '''
    def __init__(self, q, m, t0, pos0, vel0):
        self.q = q
        self.m = m
        self.pos = [pos0,]
        self.vel = [vel0,]
        self.time = [t0,]
        self.alive = True

    def addTime(self, t):
        self.time.append(t)

    def addPos(self, pos):
        self.pos.append(pos)

    def addVel(self, vel):
        self.vel.append(vel)

    def reset(self):
        t0 = self.time[0]
        p0 = self.pos[0]
        v0 = self.vel[0]
        self.t0 = [t0, ]
        self.pos = [p0, ]
        self.vel = [v0, ]

def unitSphereVector(N):
    '''
    Sample N points uniformly on a sphere of radius 1
    '''
    phi = 2*np.pi*np.random.uniform(size=N)
    z = 1-2*np.random.uniform(size=N)
    theta = np.arccos(z)
    return np.array([np.sin(theta)*np.cos(phi), np.sin(theta)*np.sin(phi), z]).T

def initializeChargeCloud(q, m, N, t0, R, R0):
    '''
    Create a list of Quasiparticle objects along a path smeared with a gaussian distribution
    :param q: charge in Coulomb of individual quasiparticle
    :param m: mass in kg of individual quasiparticle
    :param N: number of quasiparticles to create
    :param R: Radius of gaussian smoothing (sigma)
    :param R0: either (DIMENSIONS) array or (N, DIMENSIONS) array describing initial positions
    '''
    pos = R0 + R*np.random.normal(size=(N, 3))
    #velS = maxwell.rvs(size=N, scale=scale)
    #TODO fix scale for Maxwell
    velS = np.zeros(N)
    velDir = unitSphereVector(N)
    
    cc = []
    for i in range(N):
        cc.append(Quasiparticle(q, m, t0, pos[i], velS[i]*velDir[i]))
    return cc



########################################
# Physical Constants
########################################

# Standard energy scale is eV
# Standard length scale is nm
# Standard temperature scale is K
# Constants in SI units have _SI in name

me = 511e3 #electron mass in eV/c2
bohr_a0 = 0.0529 # Bohr length in nm
hartree = 27.211 # eV
e2 = 1.4399 # Elemental charge squared in eV nm
alpha = 1/137
hbarc = 0.197326e3 # eV nm
me_SI = 9e-31 # electron mass in kg
qe_SI = 1.6e-19 # electron charge in C
kB_SI = 1.38e-23 # Boltzmann constant in SI units
kB = 8.617e-5 # boltzmann constant in eV/K
kE_SI = 8.99e9 # 1/(4 pi epsilon0) in SI units

#######################################
# Material Constants
#######################################

epsR_Si = 11.8 # Relative dielectric strength of Silicon
E0gap_Si = 1.1692 # eV
Fano_Si = 0.11

DIMENSIONS=3



from scipy.interpolate import RegularGridInterpolator 
from .interp_3d import *
import numpy as np

class Potential:
    '''
    The Potential class is used to create electric and weighting potential objects. 
    These can then be converted to Field objects for simulation. 
    We use Potential objects mainly for plotting and transferring field information between nessie and our other field solving tools. 
    data is assumed to be a 3-dimensional array. grid is an array of axes (x,y,z) where each axis is an array of positions in meters. 
    '''
    def __init__(self, _name, _data=None, _grid=None):
        self.name = _name
        self.data = _data
        self.grid = _grid

    def __str__(self):
        return ("Nessie Potential object \n Name: %s \n Size: %s \n" 
                % (self.name, np.shape(self.data)))

class Field:
    '''
    The Field class is used to create electric and weighting field objects. 
    We use cartesian coordinates and define each field component as its own 3-dimensional array (fieldx, fieldy, fieldz). 
    grid is the same as in the Potential class. 
    '''
    
    def __init__(self, _name, _fieldx=None, _fieldy=None,_fieldz=None, _grid=None):
        self.name = _name
        self.fieldx = _fieldx
        self.fieldy = _fieldy
        self.fieldz = _fieldz
        self.grid = _grid

    def __str__(self):
        return ("Nessie Field object \n Name: %s \n Size: %s \n"
                % (self.name, np.shape(self.fieldx)))
     
    def interpolate(self, interp3d=False):
        '''
        Returns linear interpolations of the field. 
        interp3d is a cython implementation of grid interpolation, we have modified this from https://github.com/jglaser/interp3d to work for non-regular grids. 
        If interp3d is set to False instead the scipy RegularGridInterpolator will be used. 
        This is a pure python interpolator and is significantly slower. 
        Each interpolator takes input differently so which is used is tracked throughout the simulation.  
        '''
        fieldShape = np.shape(self.fieldx)
        x = self.grid[0]
        y = self.grid[1]
        z = self.grid[2]
        
        if interp3d:
            
            fieldx_interp = Interp3D(self.fieldx.astype('double', order="C"), x,y,z)
            fieldy_interp = Interp3D(self.fieldy.astype('double',order="C"), x,y,z)
            fieldz_interp = Interp3D(self.fieldz.astype('double',order="C"), x,y,z)
            
            fieldMag = np.sqrt(self.fieldx**2+self.fieldy**2+self.fieldz**2)
            fieldMag_interp = Interp3D(fieldMag.astype('double',order="C"), x,y,z)
            
        else:   
            fieldx_interp = RegularGridInterpolator((x,y,z),self.fieldx)
            fieldy_interp = RegularGridInterpolator((x,y,z),self.fieldy)
            fieldz_interp = RegularGridInterpolator((x,y,z),self.fieldz)
            fieldMag = np.sqrt(self.fieldx**2+self.fieldy**2+self.fieldz**2)
            fieldMag_interp = RegularGridInterpolator((x,y,z),fieldMag)
        
        return fieldx_interp, fieldy_interp, fieldz_interp, fieldMag_interp


def weightingPotentialFromH5(filename, rotate90=True):
    '''
    Import weighting potential from hdf5 file. 
    When saving files to hdf5 from Julia a rotation is applied to the array, so when importing from SSD we need to rotate back.
    '''
    import h5py
    f = h5py.File(filename, 'r')
    data = np.array(f["wp"])
    if rotate90: data = np.rot90(data,axes=(0,2))
    gridx = np.array(f["gridx"])
    gridy = np.array(f["gridy"])
    gridz = np.array(f["gridz"])
    grid = [gridx,gridy,gridz]
    name = filename[:-3]
    weightingPotential = Potential(name, data, grid)
    return weightingPotential

#Import electric field from hdf5 file.
def eFieldFromH5(filename, rotate90=True):
    import h5py
    f = h5py.File(filename, 'r')
    
    if rotate90:
        Ex = np.rot90(np.array(f["Ex"]),axes=(0,2))
        Ey = np.rot90(np.array(f["Ey"]),axes=(0,2))
        Ez = np.rot90(np.array(f["Ez"]),axes=(0,2))
    else:
        Ex = np.array(f["Ex"])
        Ey = np.array(f["Ey"])
        Ez = np.array(f["Ez"])
        
    gridx = np.array(f["gridx"])
    gridy = np.array(f["gridy"])
    gridz = np.array(f["gridz"])
    grid = [gridx,gridy,gridz]
    name = filename[:-3]

    Efield = Field(name, Ex, Ey,Ez, grid)
    
    return Efield



cimport numpy as np
import numpy as np
from libc.math cimport floor
from cython cimport boundscheck, wraparound, nonecheck, cdivision

@cdivision(True)
cpdef np.float_t _interp3D(np.float_t[:,:,::1] v, np.float_t x, np.float_t y, np.float_t z, int X, int Y, int Z):

    cdef:
        int i, x0, x1, y0, y1, z0, z1, dim
        np.float_t xd, yd, zd, c00, c01, c10, c11, c0, c1, c
        np.float_t *v_c

    v_c = &v[0,0,0]

    x0 = <int>floor(x)
    x1 = x0 + 1
    y0 = <int>floor(y)
    y1 = y0 + 1
    z0 = <int>floor(z)
    z1 = z0 + 1

    xd = (x-x0)/(x1-x0)
    yd = (y-y0)/(y1-y0)
    zd = (z-z0)/(z1-z0)

    if x0 >= 0 and y0 >= 0 and z0 >= 0 and x1 < X and y1 < Y and z1 < Z:
        c00 = v_c[Y*Z*x0+Z*y0+z0]*(1-xd) + v_c[Y*Z*x1+Z*y0+z0]*xd
        c01 = v_c[Y*Z*x0+Z*y0+z1]*(1-xd) + v_c[Y*Z*x1+Z*y0+z1]*xd
        c10 = v_c[Y*Z*x0+Z*y1+z0]*(1-xd) + v_c[Y*Z*x1+Z*y1+z0]*xd
        c11 = v_c[Y*Z*x0+Z*y1+z1]*(1-xd) + v_c[Y*Z*x1+Z*y1+z1]*xd

        c0 = c00*(1-yd) + c10*yd
        c1 = c01*(1-yd) + c11*yd

        c = c0*(1-zd) + c1*zd

    else:
        c = 0
    return c



import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import matplotlib.animation as animation
import numpy as np
from scipy.interpolate import RegularGridInterpolator 

plt.rcParams['figure.figsize'] = [8, 5]
plt.rcParams.update({'font.size': 18})

def plot_current(event, show_plot=True, alpha=1):
    plt.plot(event.dt,event.dI, alpha=alpha)
    if show_plot: plt.show()
    return None
    
def plot_signal(event, show_plot=True, alpha=1):
    plt.plot(event.signal_times,event.signal_I, alpha=alpha)
    if show_plot: plt.show()
    return None


def plot_event_drift(event, bounds, prefix="",suffix="", show_plot=True, pairs=None, save=False):
    fig = plt.figure()
    ax = plt.axes(projection ='3d')
    
    N = len(event.quasiparticles) if pairs == None else 2*pairs
    
    try:
        for i in range(N):
            pos = np.array(event.quasiparticles[i].pos)

            if event.quasiparticles[i].q < 0:
	            ax.plot3D(pos[:, 0], pos[:, 1], pos[:, 2], 'green')
            else:
                ax.plot3D(pos[:, 0], pos[:, 1], pos[:, 2], 'red')
    except: 
        print("No electron drift paths")
	
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.set_xlim(*bounds[0])
    ax.set_ylim(*bounds[1])
    ax.set_zlim(*bounds[2])
    
    if show_plot: plt.show()
    
    if save: plt.savefig(prefix+"drift_path" +suffix + ".png")
    return None
    
def plot_field_lines(field, mesh_size = (500,500), x_plane = True, density = 2,
                        prefix="",suffix="", show_plot=True, log=True, bounds=None):
                        
    fieldx_interp, fieldy_interp, fieldz_interp, fieldMag_interp = field.interpolate()
    
    if bounds is None:
        bounds = [[axis[0],axis[-1]] for axis in field.grid]
    
    ni, nj = mesh_size[0], mesh_size[1]
    
    if x_plane:
        i = np.linspace(bounds[1][0], bounds[1][1], ni)
        j = np.linspace(bounds[2][0], bounds[2][1], nj)
        Y, Z = np.meshgrid(i,j)
        X = np.zeros((ni,nj))
        coords = np.stack((X,Y,Z),axis=-1)
        Ex, Ey, Ez = fieldy_interp(coords), fieldz_interp(coords), fieldz_interp(coords)
    else:
        i = np.linspace(bounds[0][0], bounds[0][1], ni)
        j = np.linspace(bounds[2][0], bounds[2][1], nj)
        X, Z = np.meshgrid(i,j)
        Y = np.zeros((ni,nj))
        coords = np.stack((X,Y,Z),axis=-1)
        Ex, Ey, Ez = fieldy_interp(coords),fieldy_interp(coords), fieldz_interp(coords)
    
    color = np.sqrt(Ex**2+Ey**2+Ez**2) if log else Ez
    
    fig, ax = plt.subplots()
    
    strm = ax.streamplot(i, j, Ex, Ez, color=color, linewidth=1, cmap=plt.cm.inferno,
              density=density, arrowstyle='->', arrowsize=1.5)
    fig.colorbar(strm.lines, label="Electric Field Magnitude (V/m)")
              
    xlabel = "y" if x_plane else "x"          
    ax.set_xlabel(xlabel) 
    ax.set_ylabel('z')
    ax.set_xlim(i[0],i[-1])
    ax.set_ylim(j[0],j[-1])
    
    if show_plot: plt.show()
    
    plt.savefig(prefix+"field" +suffix + ".png")
    
    return None
    
def plot_potential(potential, mesh_size = (500,500), x_plane = True,
                        prefix="",suffix="", show_plot=True, bounds=None):
                            
    x = potential.grid[0]
    y = potential.grid[1]
    z = potential.grid[2]
    
    if bounds is None:
        bounds = [[axis[0],axis[-1]] for axis in potential.grid]
           
    potential_interp  = RegularGridInterpolator((x,y,z),potential.data)
    
    ni, nj = mesh_size[0], mesh_size[1]
    
    if x_plane:
        i = np.linspace(bounds[1][0], bounds[1][1], ni)
        j = np.linspace(bounds[2][0], bounds[2][1], nj)
        Y, Z = np.meshgrid(i,j)
        X = np.zeros((ni,nj))
        coords = np.stack((X,Y,Z),axis=-1)
        potentialGrid = potential_interp(coords)
        plt.contourf(Y,Z,potentialGrid,mesh_size[0])
    else:
        i = np.linspace(bounds[0][0], bounds[0][1], ni)
        j = np.linspace(bounds[2][0], bounds[2][1], nj)
        X, Z = np.meshgrid(i,j)
        Y = np.zeros((ni,nj))
        coords = np.stack((X,Y,Z),axis=-1)
        potentialGrid = potential_interp(coords)
        plt.contourf(X,Z,potentialGrid,mesh_size[0])
    
    if show_plot: plt.show()
    
    plt.savefig(prefix+"potential" +suffix + ".png")
    
    return None



import numpy as np
import pyximport
pyximport.install(setup_args={'include_dirs': [np.get_include()]})

from numba import jit

from .interp import _interp3D

@jit(nopython=True)
def find_first(item, vec):
    for i, val in enumerate(vec):
        if item < val:
            return i
    return -1

@jit(nopython=True)
def bisect_left(a, x):
    hi = len(a)
    lo = 0
    while lo < hi:
        mid = (lo + hi) // 2
        if a[mid] < x:
            lo = mid + 1
        else:
            hi = mid
    return lo
 
@jit(nopython=True)
def get_ijk(t, x, y, z):
    i_s = [0,0,0]
    axes = [x,y,z]
    for d in range(3):
        for i, val in enumerate(axes[d]):
            if t[d] < val:
                i_s[d] = i -1
                
    return i_s
    

class Interp3D(object):
    '''
    Grid interpolator for 3-dimensional regular rectangular grid. 
    This is an old version, see function below.
    '''
    def __init__(self, v, x, y, z):
        self.v = v
        self.min_x, self.max_x = x[0], x[-1]
        self.min_y, self.max_y = y[0], y[-1]
        self.min_z, self.max_z = z[0], z[-1]
        self.delta_x = (self.max_x - self.min_x)/(x.shape[0]-1)
        self.delta_y = (self.max_y - self.min_y)/(y.shape[0]-1)
        self.delta_z = (self.max_z - self.min_z)/(z.shape[0]-1)

    def __call__(self, t):
        X,Y,Z = self.v.shape[0], self.v.shape[1], self.v.shape[2]

        x = (t[0]-self.min_x)/self.delta_x
        y = (t[1]-self.min_y)/self.delta_y
        z = (t[2]-self.min_z)/self.delta_z


        return _interp3D(self.v, x, y, z, X, Y, Z)
        
        
class Interp3D(object):
    '''
    Grid interpolator for 3-dimensional rectangular grid. 
    Note that this is changed from above to work for non-regular grids.
    get_ijk and get_lmn are seperate functions mainly for testing purposes.
    '''
    def __init__(self, v, x, y, z):
        self.v = v
        self.x = x
        self.y = y
        self.z = z
        self.delta_x = np.diff(x)
        self.delta_y = np.diff(y)
        self.delta_z = np.diff(z)
        
    def get_ijk(self, t):
        return find_first(t[0], self.x)-1, find_first(t[1], self.y)-1, find_first(t[2], self.z)-1

    def get_lmn(self, t, i, j, k):
        return i + (t[0]-self.x[i])/self.delta_x[i], j + (t[1]-self.y[j])/self.delta_y[j], k + (t[2]-self.z[k])/self.delta_z[k]

    def __call__(self, t):
        X,Y,Z = self.v.shape[0], self.v.shape[1], self.v.shape[2]

        #i = np.where(self.x>t[0])[0][0]-1
        #j = np.where(self.y>t[1])[0][0]-1
        #k = np.where(self.z>t[2])[0][0]-1

        i,j,k = self.get_ijk(t)
        
        l,m,n = self.get_lmn(t, i, j, k)

        return _interp3D(self.v, l, m, n, X, Y, Z)



from .constants import *

def Egap(T):
    # Bandgap energy in eV
    a = 4.9e-4
    b = 655 # K
    return E0gap_Si - a*T*T/(T+b)

def ephCanali(T):
    return 2.15*Egap(T)+1.21

def ephBestFit(T):
    return 4.26*Egap(T)-1.15
